from __future__ import annotations

import asyncio
from datetime import timedelta
from typing import TYPE_CHECKING, Any, TypedDict

from temporalio import workflow as temporal_workflow
from temporalio.common import RetryPolicy
from temporalio.exceptions import ApplicationError
from typing_extensions import NotRequired, Unpack

from .workflow import ChildStart

if TYPE_CHECKING:
    from collections.abc import Callable
from temporalio.workflow import (
    ChildWorkflowCancellationType,
    ParentClosePolicy,
)

from .workflow import WorkflowLogger

log = WorkflowLogger()

get_external_agent_handle = (
    temporal_workflow.get_external_workflow_handle
)
agent_info = temporal_workflow.info
continue_as_new = temporal_workflow.continue_as_new
condition = temporal_workflow.wait_condition
import_functions = temporal_workflow.unsafe.imports_passed_through
uuid = temporal_workflow.uuid4

__all__ = [
    "NonRetryableError",
    "RetryPolicy",
    "RetryableError",
    "agent_info",
    "condition",
    "continue_as_new",
    "get_external_agent_handle",
    "import_functions",
    "log",
    "uuid",
]

RestackFunction = Any
RestackFunctionInput = Any


class NonRetryableError(ApplicationError):
    def __init__(self, message: str) -> None:
        super().__init__(message=message, non_retryable=True)


class RetryableError(ApplicationError):
    def __init__(self, message: str) -> None:
        super().__init__(message=message, non_retryable=False)


class StepKwargs(TypedDict):
    function: RestackFunction
    function_input: NotRequired[RestackFunctionInput]
    task_queue: NotRequired[str]
    retry_policy: NotRequired[RetryPolicy]
    schedule_to_close_timeout: NotRequired[timedelta]


class ChildKwargs(TypedDict, total=False):
    workflow: Any
    workflow_id: str
    workflow_input: Any
    agent: Any
    agent_id: str
    agent_input: Any
    task_queue: str
    cancellation_type: ChildWorkflowCancellationType
    parent_close_policy: ParentClosePolicy
    execution_timeout: timedelta


class Agent:
    def defn(self) -> Any:
        """Define an agent.

        Example:
            .. code-block:: python

                @agent.defn()
                class Agent:
                    @agent.run

        """
        return temporal_workflow.defn(sandboxed=True)

    def state(
        self,
        fn: Any,
        name: str | None = None,
        description: str | None = None,
    ) -> Any:
        """Register a method to query an agent state.

        Example:

            .. code-block:: python

                @agent.state
                def current_state(self):
                    return self.state

        Args:
            fn: The agent instance.
            name (str, optional): State name.
            description (str, optional): State description.

        Returns:
            Any: The state data.
        """
        return temporal_workflow.query(
            fn,
            name=name,
            description=description,
        )

    def event(
        self,
        function: RestackFunction,
    ) -> Any:
        """Defines an agent event.

        Example:

            .. code-block:: python

                @agent.event
                async def event_name(self, event_name: EventInput):
                    output_data = await agent.step(
                        function=Function,
                    )
                    return output_data

        Args:
            function: The agent instance.
            event (EventInput): Pydantic model for event input.

        Returns:
            Any: The data returned by the event.
        """
        name = function.__name__
        return temporal_workflow.update(function, name=name)

    def run(self, function: RestackFunction) -> Any:
        """Defines what happens when the agent runs.

        Example:

            .. code-block:: python

                @agent.defn()
                  class Agent:
                      @agent.run
                      async def run(self, agent_input: AgentInput):
                          await agent.condition(lambda: self.done)
                          return AgentOutput(data=agent_input.data)

        Args:
            function: The agent instance.
            agent_input (AgentInput): The main run method.

        Returns:
            AgentOutput: The result of the agent after its done.
        """
        import inspect
        from functools import wraps

        sig = inspect.signature(function)
        expected_params = len(sig.parameters) - 1

        @wraps(function)
        async def wrapper(*args: Any, **_kwargs: Any) -> Any:
            if expected_params == 0:
                return await function(args[0])
            if expected_params == 1:
                return await function(
                    args[0],
                    args[1] if len(args) > 1 else None,
                )
            error_message = """
                Invalid run method signature: the run method must be defined as either:
                async def run(self) -> None
                or:
                async def run(self, agent_input: AgentInput) -> None
                Please update the run method accordingly.
            """
            raise TypeError(error_message)

        return temporal_workflow.run(wrapper)

    def condition(
        self,
        fn: Callable[[], bool],
        timeout: timedelta | None = None,
    ) -> None:
        """Wait until a condition becomes true.

        Used in agent.run() to keep running until the agent is done.

        Example:

            .. code-block:: python

                await agent.condition(
                    lambda: self.done)
                )
        """
        return temporal_workflow.wait_condition(
            fn,
            timeout=timeout,
        )

    async def step(
        self,
        **kwargs: Unpack[StepKwargs],
    ) -> Any:
        """Execute a function as a step of the agent.

        Example:

        .. code-block:: python

            with import_functions():
            from src.functions.example_function import (
                example_function,
                ExampleFunctionInput,
            )
            result = await agent.step(
                function=Function,
                function_input=FunctionInput(data=input_data),
            )

        Args:
            **kwargs: Keyword arguments for the step execution.
                function (Function): The function to execute.
                function_input(Optional[FunctionInput(BaseModel)]): Input data for the function. Recommended to use a Pydantic model for input validation.
                task_queue (str, optional): Task queue on which the function is executed.
                retry_policy (Optional[RetryPolicy]): Configure how the function should be retried when failing.
                start_to_close_timeout (Optional[timedelta]): Force function to timeout within the given duration. Default 5 minutes.

        Returns:
            FunctionOutput: The result of the function.

        """
        function = kwargs.get("function")
        function_input = kwargs.get("function_input")
        task_queue = kwargs.get("task_queue", "restack")
        retry_policy = kwargs.get("retry_policy")
        schedule_to_close_timeout = kwargs.get(
            "schedule_to_close_timeout",
        )
        schedule_to_start_timeout = kwargs.get(
            "schedule_to_start_timeout",
        )
        start_to_close_timeout = kwargs.get(
            "start_to_close_timeout",
            timedelta(minutes=5),
        )
        heartbeat_timeout = kwargs.get("heartbeat_timeout")
        engine_id = self._get_engine_id_from_client()
        return await temporal_workflow.execute_activity(
            activity=function,
            args=(function_input,)
            if function_input is not None
            else (),
            task_queue=f"{engine_id}-{task_queue}",
            start_to_close_timeout=start_to_close_timeout,
            schedule_to_close_timeout=schedule_to_close_timeout,
            schedule_to_start_timeout=schedule_to_start_timeout,
            heartbeat_timeout=heartbeat_timeout,
            retry_policy=retry_policy,
        )

    async def child_start(
        self,
        **kwargs: Unpack[ChildKwargs],
    ) -> ChildStart:
        """Start a child agent or workflow.

        Example:

        .. code-block:: python

            from .workflow.child import ChildWorkflow, ChildWorkflowInput
            from .agent.child import ChildAgent, ChildAgentInput

            workflow_result = await agent.child_start(
                workflow=ChildWorkflow,
                workflow_id="child-workflow-123",
                workflow_input=WorkflowInput(data="example"),
            )
            agent_result = await agent.child_start(
                agent=ChildAgent,
                agent_id="child-agent-123",
                agent_input=AgentInput(data="example"),
            )

        Args(required):
            # For child workflows
            workflow: The child workflow to start.
            workflow_id: A unique identifier for the child workflow.
            workflow_input: The input data for the child workflow. Pydantic model recommended.
            # For child agents
            agent: The child agent to start.
            agent_id: A unique identifier for the child agent.
            agent_input: The input data for the child agent. Pydantic model recommended.

        Args(optional):
            task_queue (str): The task queue for executing the child.
            parent_close_policy (ParentClosePolicy): Specifies the behavior when the parent is closed.
            execution_timeout (timedelta): The maximum duration allowed for the child to complete.

        Returns:
            id: The ID of the child.
            run_id: The ID of the first run of the child.

        """
        workflow = kwargs.get("workflow")
        workflow_input = kwargs.get("workflow_input")
        workflow_id = kwargs.get("workflow_id")
        agent = kwargs.get("agent")
        agent_input = kwargs.get("agent_input")
        agent_id = kwargs.get("agent_id")
        task_queue = kwargs.get("task_queue", "restack")
        cancellation_type = kwargs.get(
            "cancellation_type",
            ChildWorkflowCancellationType.WAIT_CANCELLATION_COMPLETED,
        )
        parent_close_policy = kwargs.get(
            "parent_close_policy",
            ParentClosePolicy.TERMINATE,
        )
        execution_timeout = kwargs.get("execution_timeout")

        if not workflow and not agent:
            error_message = (
                "Either workflow or agent must be provided."
            )
            log.error(error_message)
            raise ValueError(error_message)
        if workflow and agent:
            error_message = "Either workflow or agent must be provided, not both."
            log.error(error_message)
            raise ValueError(error_message)

        engine_id = self._get_engine_id_from_client()
        is_agent = workflow is None
        handle = await temporal_workflow.start_child_workflow(
            workflow=workflow or agent,
            args=[workflow_input or agent_input]
            if (workflow_input or agent_input)
            else [],
            id=self._add_engine_id_prefix(
                engine_id,
                workflow_id or agent_id,
            ),
            task_queue=f"{engine_id}-{task_queue}",
            memo={"engineId": engine_id, "agent": is_agent},
            search_attributes={"engineId": [engine_id]},
            cancellation_type=cancellation_type,
            parent_close_policy=parent_close_policy,
            execution_timeout=execution_timeout,
        )
        return ChildStart(
            id=handle.id,
            run_id=handle.first_execution_run_id,
        )

    async def child_execute(
        self,
        **kwargs: Unpack[ChildKwargs],
    ) -> Any:
        """Start a child workflow or agent and wait for it to complete.

        Example:

        .. code-block:: python

            from .workflow.child import ChildWorkflow, ChildWorkflowInput
            from .agent.child import ChildAgent, ChildAgentInput

            workflow_result = await agent.child_execute(
                workflow=ChildWorkflow,
                workflow_id="child-workflow-123",
                workflow_input=ChildWorkflowInput(data="example"),
            )
            agent_result = await agent.child_execute(
                agent=ChildAgent,
                agent_id="child-agent-123",
                agent_input=ChildAgentInput(data="example"),
            )

        Args(required):
            # For a child workflow
            workflow: The child workflow class to execute.
            workflow_id: A unique identifier for the child workflow.
            workflow_input: The input data for the child workflow. Pydantic model recommended.
            # For a child agent
            agent: The child agent class to execute.
            agent_id: A unique identifier for the child agent.
            agent_input: The input data for the child agent. Pydantic model recommended.

        Args(optional):
            task_queue (str): The task queue for executing the child.
            parent_close_policy (ParentClosePolicy): Specifies the behavior when the parent is closed.
            execution_timeout (timedelta): The maximum duration allowed for the child to complete.

        Returns:
            Any: The result of the child workflow or agent.

        """
        workflow = kwargs.get("workflow")
        workflow_input = kwargs.get("workflow_input")
        workflow_id = kwargs.get("workflow_id")
        agent = kwargs.get("agent")
        agent_input = kwargs.get("agent_input")
        agent_id = kwargs.get("agent_id")
        task_queue = kwargs.get("task_queue", "restack")
        cancellation_type = kwargs.get(
            "cancellation_type",
            ChildWorkflowCancellationType.WAIT_CANCELLATION_COMPLETED,
        )
        parent_close_policy = kwargs.get(
            "parent_close_policy",
            ParentClosePolicy.TERMINATE,
        )
        execution_timeout = kwargs.get("execution_timeout")
        if not workflow and not agent:
            error_message = (
                "Either workflow or agent must be provided."
            )
            log.error(error_message)
            raise ValueError(error_message)
        if workflow and agent:
            error_message = "Either workflow or agent must be provided, not both."
            log.error(error_message)
            raise ValueError(error_message)

        engine_id = self._get_engine_id_from_client()

        is_agent = workflow is None
        return await temporal_workflow.execute_child_workflow(
            workflow=workflow or agent,
            args=[workflow_input or agent_input]
            if (workflow_input or agent_input)
            else [],
            id=self._add_engine_id_prefix(
                engine_id,
                workflow_id or agent_id,
            ),
            task_queue=f"{engine_id}-{task_queue}",
            memo={"engineId": engine_id, "agent": is_agent},
            search_attributes={"engineId": [engine_id]},
            cancellation_type=cancellation_type,
            parent_close_policy=parent_close_policy,
            execution_timeout=execution_timeout,
        )

    async def sleep(self, seconds: int) -> Any:
        """Waits for a set duration.

        Example:

            .. code-block:: python
                await agent.sleep(5)

        Args:
            seconds (int): Time to sleep in seconds.

        Returns:
            None
        """
        return await asyncio.sleep(seconds)

    def _get_engine_id_from_client(self) -> Any:
        return temporal_workflow.memo_value("engineId", "local")

    def _add_engine_id_prefix(
        self,
        engine_id: str,
        agent_id: str,
    ) -> str:
        if agent_id.startswith(f"{engine_id}-"):
            return agent_id
        return f"{engine_id}-{agent_id}"


agent = Agent()



"""Defines the PlaygroundRun workflow for executing activities in a Temporal workflow.

This module contains:
- The `PlaygroundInput` data class for workflow input data.
- The `PlaygroundRun` workflow definition that executes an activity.

"""



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class AgentEvent:
    name: str
    input: dict[str, Any] | None = None


@dataclass
class SendAgentEvent:
    event: AgentEvent
    workflow: str | None = None



# ruff: noqa: N815
# PlaygroundRun needs consistent typing between frontend, server, typescript and python library

from dataclasses import dataclass
from datetime import timedelta
from typing import Any

from temporalio import workflow


@dataclass
class PlaygroundInput:
    functionName: str
    taskQueue: str
    input: Any | None = None


@workflow.defn(name="playgroundRun", sandboxed=True)
class PlaygroundRun:
    @workflow.run
    async def run(self, params: PlaygroundInput) -> Any:
        engine_id = workflow.memo_value("engineId", "local")
        return await workflow.execute_activity(
            activity=params.functionName,
            task_queue=f"{engine_id}-{params.taskQueue}",
            args=[params.input],
            start_to_close_timeout=timedelta(seconds=120),
        )


@workflow.defn()
class FunctionsOnly:
    """FunctionsOnly workflow that prevents TaskLocals issues for functions-only services.

    This workflow should never actually be executed - it exists only to
    ensure that workers have at least one workflow registered.
    """

    @workflow.run
    async def run(self) -> str:
        # This should never be called
        return "functions_only_workflow_should_not_be_called"



from __future__ import annotations

import asyncio
import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypedDict

import aiohttp
from temporalio.client import (
    Client,
    Schedule,
    ScheduleActionStartWorkflow,
    ScheduleCalendarSpec,
    ScheduleIntervalSpec,
    ScheduleRange,
    ScheduleSpec,
    WithStartWorkflowOperation,
)
from temporalio.common import (
    SearchAttributeKey,
    SearchAttributePair,
    TypedSearchAttributes,
    WorkflowIDConflictPolicy,
)
from temporalio.contrib.pydantic import pydantic_data_converter
from temporalio.runtime import (
    OpenTelemetryConfig,
    Runtime,
    TelemetryConfig,
)
from temporalio.worker import (
    PollerBehaviorAutoscaling,
    ResourceBasedSlotConfig,
    ResourceBasedSlotSupplier,
    ResourceBasedTunerConfig,
    Worker,
    WorkerTuner,
)
from typing_extensions import Unpack

from .endpoints import explore_class_details
from .observability import log_with_context
from .playground.workflow import FunctionsOnly, PlaygroundRun

if TYPE_CHECKING:
    from .event import AgentEvent
    from .security import DataConverter

__all__ = [
    "ScheduleCalendarSpec",
    "ScheduleIntervalSpec",
    "ScheduleRange",
]


@dataclass
class CloudConnectionOptions:
    engine_id: str
    api_key: str
    address: str | None = "localhost:7233"
    api_address: str | None = None
    temporal_namespace: str | None = "default"
    data_converter: DataConverter | None = None


@dataclass
class ServiceOptions:
    rate_limit: int | None = 100000
    max_concurrent_workflow_runs: int | None = 3000
    max_concurrent_function_runs: int | None = 1000
    endpoints: bool | None = True
    endpoint_group: str | None = None


@dataclass
class ResourceOptions:
    slots: int | None = 10
    target_cpu_usage: float | None = 0.8
    target_memory_usage: float | None = 0.8


class Restack:
    """Restack client.

    Example:

    .. code-block:: python
        engine_id = os.getenv("RESTACK_ENGINE_ID")
        address = os.getenv("RESTACK_ENGINE_ADDRESS")
        api_key = os.getenv("RESTACK_ENGINE_API_KEY")
        api_address = os.getenv("RESTACK_ENGINE_API_ADDRESS")

        connection_options = CloudConnectionOptions(
            engine_id=engine_id,
            address=address,
            api_key=api_key,
            api_address=api_address,
        )
        client = Restack(connection_options)
    """

    api_key: str | None = None
    client: Client | None = None
    options: CloudConnectionOptions | None = None
    functions: list[Any] | None = None

    def __init__(
        self,
        options: CloudConnectionOptions | None = None,
    ) -> None:
        super().__init__()
        self.client = None
        self.options = options

    def get_connection_options(self) -> dict[str, Any]:
        target_host = (
            self.options.address
            if self.options and self.options.address is not None
            else "localhost:7233"
        )
        api_address = (
            f"https://{self.options.api_address}"
            if self.options
            and self.options.api_address is not None
            else "http://localhost:6233"
        )
        engine_id = (
            self.options.engine_id
            if self.options and self.options.engine_id is not None
            else "local"
        )
        options = {
            "target_host": target_host,
            "metadata": {
                "restack-engineId": engine_id,
                "restack-apiAddress": api_address,
            },
        }
        if self.options and self.options.api_key is not None:
            options["tls"] = True
            options["api_key"] = self.options.api_key
        return options

    async def connect(
        self,
        connection_options: CloudConnectionOptions | None = None,
    ) -> None:
        """Connect to the Restack Engine.

        Args:
            connection_options (CloudConnectionOptions | None): Optional connection options.

        """
        if self.client:
            log_with_context(
                "INFO",
                "Reusing existing client connection",
            )
            return
        try:
            self.client = await self.create_client(
                connection_options,
            )
            log_with_context(
                "INFO",
                "Connected to Restack Engine",
            )
        except Exception as e:
            log_with_context(
                "ERROR",
                "Failed to connect to Restack Engine",
                error=str(e),
            )
            raise

    async def create_client(
        self,
        connection_options: CloudConnectionOptions | None = None,
    ) -> Client:
        connect_options = (
            connection_options or self.get_connection_options()
        )

        target_host = connect_options["target_host"]
        namespace = (
            self.options.temporal_namespace
            if self.options
            and self.options.temporal_namespace is not None
            else "default"
        )
        api_key = (
            self.options.api_key
            if self.options and self.options.api_key is not None
            else None
        )
        tls = bool(self.options and self.options.api_key)
        metadata = connect_options["metadata"]
        data_converter = (
            self.options.data_converter
            if self.options
            and self.options.data_converter is not None
            else pydantic_data_converter
        )

        if (
            os.getenv("RESTACK_TELEMETRY_OTEL_ADDRESS")
            is not None
        ):
            runtime = Runtime(
                telemetry=TelemetryConfig(
                    metrics=OpenTelemetryConfig(
                        url=os.getenv(
                            "RESTACK_TELEMETRY_OTEL_ADDRESS",
                        ),
                    ),
                ),
            )
        else:
            runtime = Runtime(telemetry=TelemetryConfig())

        try:
            client = await Client.connect(
                target_host,
                namespace=namespace,
                api_key=api_key,
                tls=tls,
                rpc_metadata=metadata,
                data_converter=data_converter,
                runtime=runtime,
            )
        except Exception as e:
            documentation_url = "https://docs.restack.io/help/troubleshooting#restack-is-not-running"
            log_with_context(
                "ERROR",
                (
                    "Failed to connect, make sure Restack is running: "
                    + documentation_url
                ),
                error=str(e),
            )
            raise
        return client

    async def create_service(  # noqa: PLR0913
        self,
        workflows: list[Any] | None = None,
        agents: list[Any] | None = None,
        functions: list[Any] | None = None,
        task_queue: str | None = None,
        options: ServiceOptions | None = None,
        resources: ResourceOptions | None = None,
    ) -> Worker:
        try:
            log_with_context(
                "INFO",
                f"Starting service on task queue {task_queue}",
            )
            client = await self.create_client()
            engine_id = self.get_connection_options()["metadata"][
                "restack-engineId"
            ]
            api_address = self.get_connection_options()[
                "metadata"
            ]["restack-apiAddress"]
            identity = f"{engine_id}-{os.getpid()}"
            task_queue = f"{engine_id}-{task_queue or 'restack'}"

            all_workflows = (workflows or []) + (agents or [])

            if options is None:
                options = ServiceOptions()

            if resources:
                resource_based_options = ResourceBasedTunerConfig(
                    resources.target_cpu_usage or 0.8,
                    resources.target_memory_usage or 0.8,
                )
                tuner = WorkerTuner.create_composite(
                    workflow_supplier=ResourceBasedSlotSupplier(
                        ResourceBasedSlotConfig(),
                        resource_based_options,
                    ),
                    activity_supplier=ResourceBasedSlotSupplier(
                        ResourceBasedSlotConfig(),
                        resource_based_options,
                    ),
                    local_activity_supplier=ResourceBasedSlotSupplier(
                        ResourceBasedSlotConfig(),
                        resource_based_options,
                    ),
                )

                service = Worker(
                    identity=identity,
                    client=client,
                    task_queue=task_queue,
                    workflows=all_workflows,
                    activities=functions or [],
                    max_task_queue_activities_per_second=options.rate_limit,
                    tuner=tuner,
                    workflow_task_poller_behavior=PollerBehaviorAutoscaling(),
                    activity_task_poller_behavior=PollerBehaviorAutoscaling(),
                )
            else:
                service = Worker(
                    identity=identity,
                    client=client,
                    task_queue=task_queue,
                    workflows=all_workflows,
                    activities=functions or [],
                    max_task_queue_activities_per_second=options.rate_limit,
                    max_concurrent_activities=options.max_concurrent_function_runs,
                    max_concurrent_workflow_tasks=options.max_concurrent_workflow_runs,
                )
        except Exception as e:
            if (
                e.__cause__
                and "http.client.incompleteread.__mro_entries__".lower()
                in str(e.__cause__).lower()
            ):
                log_with_context(
                    "ERROR",
                    "Failed to start service: Functions in workflow or agents steps need to be imported with import_functions(). See docs at https://docs.restack.io/libraries/python/workflows",
                )
            else:
                log_with_context(
                    "ERROR",
                    "Failed to start service",
                    error=str(e),
                )
            raise
        else:
            if options.endpoints:
                async with aiohttp.ClientSession() as session:
                    workflow_details = []
                    for workflow in workflows or []:
                        details = explore_class_details(workflow)
                        details["task_queue"] = (
                            task_queue.replace(
                                f"{engine_id}-",
                                "",
                            )
                        )
                        workflow_details.append(details)

                    agent_details = []
                    for agent in agents or []:
                        details = explore_class_details(agent)
                        details["task_queue"] = (
                            task_queue.replace(
                                f"{engine_id}-",
                                "",
                            )
                        )
                        agent_details.append(details)

                    endpoint_identity = (
                        f"{engine_id}-{options.endpoint_group or os.getpid()}"
                        if engine_id == "local"
                        else f"{engine_id}"
                    )
                    endpoints_payload = {
                        "identity": endpoint_identity,
                        "workflows": workflow_details,
                        "agents": agent_details,
                    }

                    try:
                        async with session.post(
                            f"{api_address}/api/engine/endpoints",
                            json=endpoints_payload,
                        ) as response:
                            response_text = await response.text()
                            log_with_context(
                                "INFO",
                                f"{response_text}",
                            )
                    except Exception as e:
                        log_with_context(
                            "WARNING",
                            "Failed to register endpoints",
                            error=str(e),
                        )
                        raise
            return service

    async def run_service(self, service: Worker) -> None:
        try:
            engine_id = self.get_connection_options()["metadata"][
                "restack-engineId"
            ]

            log_task_queue = service.task_queue.replace(
                f"{engine_id}-",
                "",
            )

            if service.task_queue.endswith("restack"):
                playground_service = Worker(
                    identity=f"{engine_id}-playground-{os.getpid()}",
                    client=service.client,
                    task_queue=f"{engine_id}-playground",
                    workflows=[PlaygroundRun],
                )

                log_with_context(
                    "INFO",
                    f"Service on task queue {log_task_queue} ready",
                )
                await asyncio.gather(
                    service.run(),
                    playground_service.run(),
                )
            else:
                log_with_context(
                    "INFO",
                    f"Service on task queue {log_task_queue} ready",
                )

                await service.run()
        except Exception as e:
            log_with_context(
                "ERROR",
                "Failed to run service",
                error=str(e),
            )
            raise

    class StartServiceKwargs(TypedDict):
        workflows: list[Any]
        agents: list[Any]
        functions: list[Any]
        task_queue: str
        options: ServiceOptions

    async def start_service(
        self,
        **kwargs: Unpack[StartServiceKwargs],
    ) -> None:
        """Start a service with workflows, agents, or functions.

        Example:

        .. code-block:: python

            from restack_ai import Restack

            restack = Restack()
            await restack.start_service(
                workflows=[Workflow],
                agents=[Agent],
                functions=[Function],
            )

        Args:
            **kwargs (StartServiceKwargs): Keyword arguments for the start service method.
                - workflows (list[Workflow] | None): A list of workflows to start.
                - agents (list[Agent] | None): A list of agents to be used.
                - functions (list[Function] | None): A list of functions to execute.
                - task_queue (str | None): The name of the task queue.
                - options (ServiceOptions | None): Additional service options.

        """
        workflows = kwargs.get("workflows", [])
        agents = kwargs.get("agents", [])
        functions = kwargs.get("functions", [])
        task_queue = kwargs.get("task_queue", "restack")
        options = kwargs.get("options", ServiceOptions())
        resources = kwargs.get("resources")

        if not workflows and not agents and not functions:
            error_message = "At least one workflow, agent, or function must be provided"
            raise ValueError(error_message)

        if functions and not workflows and not agents:
            workflows = [FunctionsOnly]
            if options is None:
                options = ServiceOptions(endpoints=False)
            else:
                options = ServiceOptions(
                    rate_limit=options.rate_limit,
                    max_concurrent_workflow_runs=options.max_concurrent_workflow_runs,
                    max_concurrent_function_runs=options.max_concurrent_function_runs,
                    endpoints=False,
                    endpoint_group=options.endpoint_group,
                )

        service = await self.create_service(
            task_queue=task_queue,
            workflows=workflows,
            agents=agents,
            functions=functions or [],
            options=options or ServiceOptions(),
            resources=resources,
        )
        await self.run_service(service)

    class ScheduleWorkflowKwargs(TypedDict, total=False):
        task_queue: str
        schedule: ScheduleSpec
        event: AgentEvent
        is_agent: bool
        workflow_input: dict[str, Any] | None
        workflow_id: str
        workflow_name: str

    async def schedule_workflow(
        self,
        **kwargs: Unpack[ScheduleWorkflowKwargs],
    ) -> str:
        workflow_input = kwargs.get("workflow_input")
        workflow_id = kwargs.get("workflow_id")
        workflow_name = kwargs.get("workflow_name")
        task_queue = kwargs.get("task_queue", "restack")
        schedule = kwargs.get("schedule")
        event = kwargs.get("event")
        is_agent = kwargs.get("is_agent", False)
        await self.connect()
        if self.client:
            try:
                connection_options = self.get_connection_options()
                engine_id = connection_options["metadata"][
                    "restack-engineId"
                ]

                search_attribute_pairs = [
                    SearchAttributePair(
                        key=SearchAttributeKey.for_text(
                            "engineId",
                        ),
                        value=engine_id,
                    ),
                ]
                search_attributes = TypedSearchAttributes(
                    search_attributes=search_attribute_pairs,
                )

                if not schedule:
                    if event:
                        start_op = WithStartWorkflowOperation(
                            workflow_name,
                            id=f"{engine_id}-{workflow_id}",
                            id_conflict_policy=WorkflowIDConflictPolicy.USE_EXISTING,
                            args=[workflow_input]
                            if workflow_input
                            else [],
                            task_queue=f"{engine_id}-{task_queue}",
                            memo={
                                "engineId": engine_id,
                                "agent": is_agent,
                            },
                            search_attributes=search_attributes,
                        )
                        try:
                            await self.client.execute_update_with_start_workflow(
                                update=event.name,
                                args=[event.input],
                                start_workflow_operation=start_op,
                            )
                            handle = (
                                await start_op.workflow_handle()
                            )
                        except Exception as e:
                            log_with_context(
                                "ERROR",
                                "Failed to start workflow with event",
                                error=str(e),
                            )
                            raise
                        else:
                            log_with_context(
                                "INFO",
                                "Workflow started with event",
                                handle=handle,
                            )
                            return handle.first_execution_run_id

                    else:
                        handle = await self.client.start_workflow(
                            workflow_name,
                            args=[workflow_input]
                            if workflow_input
                            else [],
                            id=f"{engine_id}-{workflow_id}",
                            memo={"engineId": engine_id},
                            search_attributes=search_attributes,
                            task_queue=f"{engine_id}-{task_queue}",
                        )
                    return handle.first_execution_run_id
                schedule_action = ScheduleActionStartWorkflow(
                    workflow=workflow_name,
                    args=[workflow_input]
                    if workflow_input
                    else [],
                    id=f"{engine_id}-{workflow_id}",
                    task_queue=f"{engine_id}-{task_queue}",
                    memo={"engineId": engine_id},
                    typed_search_attributes=search_attributes,
                )
                schedule_obj = Schedule(
                    action=schedule_action,
                    spec=schedule,
                )
                scheduled = await self.client.create_schedule(
                    id=f"{engine_id}-{workflow_id}",
                    schedule=schedule_obj,
                    memo={"engineId": engine_id},
                    search_attributes=search_attributes,
                )

            except Exception as e:
                log_with_context(
                    "ERROR",
                    "Failed to start or schedule workflow",
                    error=str(e),
                )
                raise
            else:
                return scheduled.id

        else:
            error_message = "Workflow result not retrieved due to failed connection."
            log_with_context("ERROR", error_message)
            raise Exception(error_message)

    async def get_workflow_handle(
        self,
        workflow_id: str,
        run_id: str,
    ) -> Any:
        await self.connect()
        if self.client:
            try:
                connection_options = self.get_connection_options()
                engine_id = connection_options["metadata"][
                    "restack-engineId"
                ]
                if workflow_id.startswith(f"{engine_id}-"):
                    get_workflow_id = workflow_id
                else:
                    get_workflow_id = f"{engine_id}-{workflow_id}"
                return self.client.get_workflow_handle(
                    workflow_id=get_workflow_id,
                    run_id=run_id,
                )
            except Exception as e:
                log_with_context(
                    "ERROR",
                    "Failed to get workflow result",
                    error=str(e),
                )
                raise
        else:
            error_message = "Workflow result not retrieved due to failed connection."
            log_with_context("ERROR", error_message)
            raise Exception(error_message)

    async def get_workflow_result(
        self,
        workflow_id: str,
        run_id: str,
    ) -> Any:
        handle = await self.get_workflow_handle(
            workflow_id,
            run_id,
        )
        try:
            return await handle.result()
        except Exception as e:
            log_with_context(
                "ERROR",
                "Failed to get workflow result",
                error=str(e),
            )
            raise

    class ScheduleAgentKwargs(TypedDict, total=False):
        task_queue: str
        schedule: ScheduleSpec
        event: AgentEvent
        agent_input: dict[str, Any] | None
        agent_name: str
        agent_id: str

    async def schedule_agent(
        self,
        **kwargs: Unpack[ScheduleAgentKwargs],
    ) -> str:
        task_queue = kwargs.get("task_queue", "restack")
        schedule = kwargs.get("schedule")
        event = kwargs.get("event")
        agent_input = kwargs.get("agent_input")
        agent_name = kwargs.get("agent_name")
        agent_id = kwargs.get("agent_id")
        return await self.schedule_workflow(
            workflow_name=agent_name,
            workflow_id=agent_id,
            workflow_input=agent_input,
            schedule=schedule,
            task_queue=task_queue,
            event=event,
            is_agent=True,
        )

    async def get_agent_handle(
        self,
        agent_id: str,
        run_id: str,
    ) -> Any:
        return await self.get_workflow_handle(agent_id, run_id)

    async def get_agent_result(
        self,
        agent_id: str,
        run_id: str,
    ) -> Any:
        return await self.get_workflow_result(agent_id, run_id)

    async def get_agent_state(
        self,
        agent_id: str,
        run_id: str,
        state_name: str,
    ) -> Any:
        handle = await self.get_workflow_handle(agent_id, run_id)
        try:
            return await handle.query(state_name)
        except Exception as e:
            log_with_context(
                "ERROR",
                "Failed to get agent state",
                error=str(e),
            )
            raise

    async def send_agent_event(
        self,
        event_name: str,
        agent_id: str,
        run_id: str | None = None,
        event_input: dict[str, Any] | None = None,
    ) -> Any:
        handle = await self.get_workflow_handle(agent_id, run_id)
        try:
            return await handle.execute_update(
                event_name,
                event_input,
            )
        except Exception as e:
            log_with_context(
                "ERROR",
                "Failed to send agent event",
                error=str(e),
            )
            raise



import ast
import contextlib
import datetime
import logging
from typing import Any, ClassVar

import colorama
from colorama import Fore, Style
from temporalio.runtime import (
    LogForwardingConfig,
    LoggingConfig,
    Runtime,
    TelemetryConfig,
    TelemetryFilter,
)

colorama.init()

BLACK_LIST_META = [
    "sdk_component",
    "is_local",
    "attempt",
    "namespace",
    "task_token",
    "activity_id",
    "restack",
    "client_log",
]


def extract_useful_meta(meta: dict) -> dict:
    if not meta:
        return {}

    useful_meta = {}

    # Include all keys that aren't in BLACK_LIST_META
    useful_meta = {
        key: value
        for key, value in meta.items()
        if key not in BLACK_LIST_META
    }

    # Special case: rename activityType to function
    if "activity_type" in meta:
        useful_meta["function"] = meta["activity_type"]
        del useful_meta["activity_type"]

    return useful_meta


def rename_complete_activity_or_workflow_message(
    message: str,
) -> str:
    lower_case_message = message.lower()
    if "activity" in lower_case_message:
        return lower_case_message.replace("activity", "function")

    return message


class ColoredFormatter(logging.Formatter):
    """Custom formatter adding colors and improved formatting to log messages."""

    COLORS: ClassVar[dict[str, str]] = {
        "DEBUG": Fore.RESET,
        "INFO": Fore.RESET,
        "WARNING": Fore.YELLOW,
        "ERROR": Fore.RED,
        "CRITICAL": Fore.RED + Style.BRIGHT,
        "SUCCESS": Fore.GREEN + Style.BRIGHT,
    }

    def format(self, record: logging.LogRecord) -> str:
        level_color = self.COLORS.get(record.levelname, "")
        reset_color = Style.RESET_ALL

        # Timestamps are in the user's local timezone (system default)
        timestamp = (
            datetime.datetime.fromtimestamp(  # noqa: DTZ006
                record.created,
            ).strftime(
                "%Y-%m-%dT%H:%M:%S.%f",
            )[:-3]
            + "Z"
        )

        # Get the base message
        message = record.getMessage()
        # Split the message and the dictionary if present
        base_message = message
        context_dict = None

        if "({" in message and "})" in message:
            parts = message.split("({", 1)
            base_message = parts[0].strip()

            with contextlib.suppress(SyntaxError, ValueError):
                context_dict = ast.literal_eval("({" + parts[1])

        if (
            not hasattr(record, "extra_fields")
            and context_dict is None
        ):
            return ""

        if record.levelname == "DEBUG" or (
            record.levelname == "WARNING"
            and not getattr(record, "extra_fields", {}).get(
                "restack",
            )
            and not getattr(record, "extra_fields", {}).get(
                "client_log",
            )
        ):
            base_message = (
                rename_complete_activity_or_workflow_message(
                    base_message,
                )
            )

        # Start with timestamp and base message
        formatted = f"[restack] {Fore.MAGENTA}{timestamp}{reset_color} [{level_color}{record.levelname}{reset_color}] {base_message}"

        # Add context dictionary items if present

        context_dict_with_meta = extract_useful_meta(context_dict)

        if context_dict_with_meta:
            formatted += "\n"
            for key, value in context_dict_with_meta.items():
                formatted += (
                    f"  {key}: {Fore.GREEN}{value}{reset_color}\n"
                )

        # Add any additional extra fields
        if hasattr(record, "extra_fields"):
            if (
                not context_dict_with_meta
            ):  # Only add newline if not already
                formatted += "\n"
            for key, value in record.extra_fields.items():
                if key not in ["restack", "client_log"]:
                    formatted += f"  {key}: {Fore.GREEN}{value}{reset_color}"

        return formatted


def setup_logger(name: str = "restack") -> logging.Logger:
    """Set up and configure the logger with colored output."""
    # Setup single restack logger that receives all logs
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(ColoredFormatter())
    console_handler.setLevel(logging.DEBUG)

    logger.handlers.clear()
    logger.addHandler(console_handler)

    temporal_logger = logging.getLogger("temporal")
    temporal_logger.setLevel(logging.DEBUG)
    temporal_logger.addHandler(console_handler)
    temporal_logger.propagate = False

    try:
        runtime = Runtime(
            telemetry=TelemetryConfig(
                logging=LoggingConfig(
                    filter=TelemetryFilter(
                        core_level="INFO",
                        other_level="INFO",
                    ),
                    forwarding=LogForwardingConfig(
                        logger=logger,
                        append_target_to_name=False,
                        prepend_target_on_message=False,
                        overwrite_log_record_time=True,
                        append_log_fields_to_message=True,
                    ),
                ),
                metrics=None,
            ),
        )
        Runtime.set_default(runtime, error_if_already_set=False)
    except Exception as e:
        error_message = f"Failed to create Runtime: {e!s}"
        logger.exception(error_message)
        raise

    return logger


logger = setup_logger()


def log_with_context(
    level: str,
    message: str,
    **kwargs: Any,
) -> None:
    """Log messages with additional context."""
    if kwargs:
        extra = {"extra_fields": {**kwargs, "restack": True}}
    else:
        extra = {"extra_fields": {"restack": True}}
    getattr(logger, level.lower())(message, extra=extra)


__all__ = ["log_with_context", "logger"]
